#ifndef __MODEL_H__                
#define __MODEL_H__

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>

#include "fpmath.h"

#if defined (MPI_ENABLED)
#include "mpi.h"
#endif

#define QQ 999999
using namespace std;


class CaseCtrlData
{
public:
    int ni, np;         // ni = num of indiv, np = dim of beta;
    gsl_matrix * mx;    //design matrix, nxp; 
    gsl_vector * vy;    //phenotype; 
    gsl_vector * delta; //prior on beta ~ N(0, delta^2);
public:
    CaseCtrlData(int, int, real *); 
    ~CaseCtrlData(); 
};
		 
double func_f(const gsl_vector * beta, void * params);
void func_df(const gsl_vector * beta, void * params, gsl_vector * df);
void func_fdf(const gsl_vector * beta, void * params, double * f, gsl_vector * df);
double Hessian(const gsl_vector * beta, void * params);
double prDM1(CaseCtrlData * pD, gsl_vector * x);
//utility functions for case/ctrl calculations. 
double probit_func_f(const gsl_vector * beta, void * params);
void probit_func_df(const gsl_vector * beta, void * params, gsl_vector * df);
void probit_func_fdf(const gsl_vector * beta, void * params, double * f, gsl_vector * df);
double probit_Hessian(const gsl_vector * beta, void * params);
double probit_prDM(CaseCtrlData * pD, gsl_vector * x);
//utility functions for case/ctrl calculations. 


double mylike_f(const gsl_vector * beta, void * params); 


typedef vector< pair< char, char> > vSNP;
typedef map<string, vSNP> HASH;

class mmSNP 
{
public:
	char A; 
	char B; 
	double countA; 
	double countB; 
	double countQ; 
public:
	mmSNP() {A=B='N'; countA=countB=countQ=0;}
	~mmSNP(){;} 
	void assign(class mmSNP s) {A=s.A; B=s.B; countA=s.countA; countB=s.countB; countQ=s.countQ;}
};   //for merge SNPs; 

class cSNP
{
private:
	string rs; 
	int flag; 
	int chr; 
	long pos;     //the physical location; 
	long index;  //the position in genotype matrix; [0 .. nLoci-1];  
	char major; 
	char minor; 
	real maf; 
public:
	cSNP(){;} 
	~cSNP(){;} 
	cSNP(string, int, long, long, pair<char, char>, real); 

	string get_rs(void) {return rs;}
	int get_chr(void) {return chr;}
	long get_pos(void) {return pos;}
	long get_index(void) {return index;}
	int get_flag(void) {return flag;} 
	            
	void set_flag(void) {flag++;}
	void set_rs(string s) {rs = s;}
	void set_chr(int s) {chr = s;}
	void set_pos(long s) {pos = s;}
	void set_index(long s) {index = s;}
};

class cGENE
{
private:	
	string gn;
	int chr;          
	long beg, end;     //if beg == end == -1; then it's a non gene; 
	
public:
	cGENE(); 
	~cGENE();
	cGENE(string gn, int chr, long beg, long end); 
	void assign(class cGENE); 
	string get_name(void) {return gn;}
	int get_chr(void) {return chr;}
	long get_beg(void) {return beg;}
	long get_end(void) {return end;}

};

class cMCMC
{
private: 
	real h;      //hetritability; 
	real p;      //snp abundance;
	real q;      //gene abundance; 
	real l;      //tuning variable; 
	real hhat; 
	real bf; 
	real like; 
	real post; 
	double sigma; 
	vector<int> sin; //positions of snps in the study; 
	vector<double> beta; 
	//sin and gin are memory efficient when trying to build an empirical distribution on hot chain. 
public: 
	cMCMC();
	~cMCMC(); 
	cMCMC(real, real, real, real, vector<int>&); 
    void assign(class cMCMC&); 
	friend class ModelnData; 
};

class ModelnData
{
public: 
	fstream fplog; // log;
	
//#if defined (MCMC)   
private:
	double m_bufsize; 
	double m_clock1, m_clock2, m_clock3, m_clock4; 
	double m_count1, m_count2, m_count3, m_count4; 

	int m_not_snp;              //if 1 then use snpmgt and ignore the range checking. note: must no missing data.  
	real ** snpmgt; 
	
	real m_maxh; 
	real m_minh; 
	int m_maxsnp; 
	int m_minsnp; 
	real m_maxlogp;
	real m_minlogp;             // p is abundance of snp; 
	
	int m_pmin; 
	int m_pmax; 
	int m_nstart; 

//	unsigned int m_maxgene; 
//	unsigned int m_mingene; 
//	real m_maxlogq; 
//	real m_minlogq;             // q is abundance of gene; 
//    real m_maxlogl; 
//	real m_minlogl;           // lam is inflation of abundance of snp when it's in a gene cover; 
//	
//	vector<int> m_si; 
//	vector<int> m_sg;       //index set of genic snps; 
//	vector<int> m_sc;       //index set of current gene covered snps; 
//	//this three sets are basic annotation of snps; 
//	
//	vector<class cSNP> vsnp; 
//	vector<class cGENE> vgene; 
//
//	map<int, int> map_snp2grp;    //break snps into different groups according to the rank of bfs.
//	multimap<string, int> mm_gene2snp; 
//	multimap<string, int> mm_snp2gene; 

	long * m_pos2rank; 
	long * m_rank2pos; 
	int m_meanrank; 
	double m_rho; 
	double * m_mix_prop; 
	gsl_ran_discrete_t * m_discrete; 
	
	
	vector<int> icase; 
	vector<int> ictrl; 
	double * m_ph; 
	double * m_xty; 

	vector<pair<long, long > > m_keymain, m_keyrecent; 
	map<pair<long, long>, double> m_mapi2xtx; 
	map<pair<long, long>, double> m_recentxtx; 
	map<int, vector<int> > m_mapi2nbr; 
	
public:
	void get_gXX(class cMCMC&, class cMCMC&, gsl_matrix *); 
	void calc_post(class cMCMC&, class cMCMC&, gsl_vector *); 
	void calc_post(class cMCMC& , gsl_matrix *); 
	void calc_hh(class cMCMC&); 
//    void calc_post_probit_laplace(class cMCMC& s, double& bfp, vector<double>& beta); 
//    void calc_post_probit_sampling(class cMCMC& s, double& bfp, vector<double>& beta); 
	void rao_blackwell(class cMCMC&, double *, gsl_vector *); 
	double propose(class cMCMC &, int); 
	double propose_snp(double, int, vector<int>&); 
    double get_truncated_gaussian(double mean, int lg0); 
	double sample_beta_gZ(class cMCMC& cur, double bpr, vector<double>& beta, gsl_vector * gZ, gsl_vector *); 
    double propose_z(gsl_vector * gZold, gsl_vector * gZnew, int); 
    double propose_z(gsl_matrix * gZold, gsl_matrix * gZnew, int); 
		
	int compatibleQ(pair<char, char> p1, pair<char, char> p2); 
	char merge_mmSNP(class mmSNP & ref, class mmSNP & s); 
	void mcmc(); 
	
//#endif
	
private:

private:
	int randSeed;               //user may specify random seed;
	int nMaxSteps;              //maximum steps of each em run;
	int nWarmSteps; 
	
/***************************************************************************  
  above for EM runs;
***************************************************************************/ 
private: 
	int nIndiv; 
	int nCohort; 				//nIndiv = nCohort + nPanel; 
	int nDip; 
    long nLoci;                  //number of snp loci, same for everyone;
	
	fstream logfile; 
	vector<string> vsRsnum; 
	vector<string> vGin;
	vector<string> vPin; 
	vector<string> vPos; 
	string fnOutput;  			//output prefix; 
	string fnRsPos; 
	vector<int> vFileIndiv;       //num of diploids in genotype files; 
	vector<int> vFilePloid;     //ploid of input genotype files. 
	  
    string fnFILE;    //a dummy filename for flexible use; 
	string fnDOC; 
	
	
	map<string, pair<char, char> > mapRs2mm;  //rs to major, minor alleles.  
	map<string, real> mapRs2maf;              //rs to minor allele freq. 
	map<string, real> mapRs2var;            //rs to var = 2*maf*(1-maf). 
	map<string, long> mapRs2pos; 
	map<string, int> mapRs2chr; 
	int nPH; 
	vector<vector<real> > vv_phval; 
	vector<vector<int> > vv_phdex;                                             
	vector<real> vsigma_a; 
	vector<real> vsigma_d; 
	vector<real> m_beta; 
	map<string, vector<real> > mapRs2beta; 
	
	int cc; //indicator for case_contrl; 
	int probit; 

	string fnGene; 
	int nGeneFlank; 

private:
	double m_sw;   //small world proportion; 
	double m_tau;  //temparture ladder; 
	int m_chains; 

	int m_silence; 
	int m_allele_coding_mode; 
	double m_exclude_maf; 
	double m_exclude_miss; 
	int m_exclude_nopos; 
	vector<int> v_reduced_num_indiv; 
public: 
	inline void Setsilence(int s) {m_silence = s;}
	inline void Set_reduced_num_indiv(int s) {v_reduced_num_indiv.push_back(s);}
	inline void Setexcludemaf(double s) {m_exclude_maf = s;}
	inline void Setexcludemiss(double s) {m_exclude_miss = s;}
	inline void Setexcludenopos(int s) {m_exclude_nopos = s;}
	inline void Setallelecoding(int s){m_allele_coding_mode = s;} 
 	void print_progress_bar(int last, char* str, int p, int total);     
private:
	
	int m_num; 
	int m_df;     //degree of freedom; 
public:
	ModelnData();
	~ModelnData();
	
	inline void Setmnotsnp(int s) {m_not_snp = s;}
	inline void Setdf(int s) {m_df = s;}
	inline void Setnum(int s) {m_num = s;}
	inline void Setsw(double s) {m_sw = s;}
	inline void Setchains(int s) {m_chains = s;}
	inline void Settau(double s) {m_tau = s;}

	inline void SetfnFILE(string s) {fnFILE.assign(s);}
	inline void SetfnDOC(string s) {fnDOC.assign(s);}
	inline int  GetnCohort(void) {return nCohort;}
	inline void Setflank(int s) {nGeneFlank = s;}
	inline int  Getflank(void) {return nGeneFlank;}
	inline void SetfnGene(string s) {fnGene.assign(s);}
	inline string GetfnGene(void) {return fnGene;}            // for gene study. 
	inline void SetOnCaseCtrl(int s) {cc = s;}
	
	inline void SetvGin(string s) {vGin.push_back(s);}
	inline void SetvPin(string s) {vPin.push_back(s);}
	inline void SetvPos(string s) {vPos.push_back(s);}
	inline void SetfnOutput(string s) {fnOutput.assign(s);}
	inline string GetfnOutput(void) {return fnOutput;}
	inline void SetfnRsPos(string s) {fnRsPos.assign(s);}

	inline void SetrandSeed(long int s){randSeed = s;}
	inline void SetnWarmSteps(int s){nWarmSteps = s;}
	inline void SetnMaxSteps(int s){nMaxSteps = s;}
	inline void SetnLoci(long s){nLoci = s;}
	inline void SetnPH(int s){nPH = s;}	

	inline int GetnMaxSteps(void) {return nMaxSteps;}
	inline int GetnWarmSteps(void) {return nWarmSteps;}
	inline int Getnum_Gin(void) {return (int) vGin.size();}
	inline int Getnum_Pin(void) {return (int) vPin.size();}
	inline long int GetrandSeed(void){return randSeed;}
	inline long GetnLoci(void){return nLoci;}   
			 							//used by fileio.cpp to allocate memory for snpGT. 
	double bf_core(real inv_va, real inv_vd, int ni, int ns, gsl_matrix * gXX, gsl_vector * gph); 
	double calc_bf(real, real *, real *, int, double&);     //for exact 
//	double calc_residual_logbf(real, double *, double *, double, int, double&);     
	double calc_residual_logbf(real, real *, real *, double, int, double&);     

	//for quatitative phenotype bf calc. 
	int  read_rs2pos(string, long, long, long);  
	int  read_bimbam_genotype(int, long, long);
	int  read_bimbam_genotype_distribution(int, int, long, long);
	int  read_bimbam_phenotype(int);

	void write_genotype(int, int);
	void write_summary(double, real **); 
	void open_log(void); 
	void close_log(void); 

	void center(int, double *); 
	void center(int, real *); 
	void center(int, gsl_vector *); 
	void normalize(int, double *); 
	void normalize(int, real *); 
	void normalize(int, gsl_vector *); 


	void set_hmin(double  s) {m_minh = s;}
	void set_hmax(double  s) {m_maxh = s;}
	void set_pmin(int s) {m_pmin = s;}
	void set_pmax(int s) {m_pmax = s;}
	void set_smin(int s) {m_minsnp = s;}
	void set_smax(int s) {m_maxsnp = s;}
	void set_nstart(int s) {m_nstart = s;}

};    

#endif

